"
 COPYRIGHT (c) 1991 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

PrinterStream subclass:#EpsonFX1PrinterStream
	 instanceVariableNames:''
	 classVariableNames:''
	 poolDictionaries:''
	 category:'Streams-External'
!

!EpsonFX1PrinterStream class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1991 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    This class defines protocol for simple text prinitng on an Epson-FX1
    (and compatible) printer. 
    It does not support multiple fonts, but knows how to print bold, italic etc.

    I cannot tell, if this is really an Epson -
    all I had to test was a Citizen 120D - its documentation claims it to be FX1 compatible.
"
!

version
    ^ '$Header: /cvs/stx/stx/libbasic2/Attic/FX1PrStr.st,v 1.13 1996-02-02 19:53:47 cg Exp $'
! !

!EpsonFX1PrinterStream class methodsFor:'queries'!

printerTypeName
    "return a descriptive name"

    ^ 'Epson FX1 compatible printer'
! !

!EpsonFX1PrinterStream methodsFor:'access writing'!

bold
    "set font to bold"

    "send: <ESC>G "

    super escape:$G
!

character:char fromCharacterSet:set
    "send a character from an alternate character set"

    "send: <ESC>R<set><char><ESC>R<0>"

    super escape:$R.
    super nextPut:(Character value:set).
    super nextPut:char.
    super escape:$R.
    super nextPut:(Character value:0)
!

cr
    "have to output cr-nl here"

    super nextPutUntranslated:(Character value:13).
    super nextPutUntranslated:(Character value:10).
    self spaces:LeftMargin
!

italic
    "set font to italic"

    "send: <ESC>4 will do that"

    super escape:$4
!

nextPut:aCharacter
    "catch special characters 
     - currently only german umlauts are handled - needs more"

    |ascii|

    ascii := aCharacter asciiValue.
    (ascii < 128) ifTrue:[
	^ super nextPut:aCharacter
    ].
    (ascii == 16rfc) ifTrue:[   "udiaeresis"
	^ self character:$} fromCharacterSet:2
    ].
    (ascii == 16re4) ifTrue:[   "adiaeresis"
	^ self character:${ fromCharacterSet:2
    ].
    (ascii == 16rf6) ifTrue:[   "odiaeresis"
	^ self character:$| fromCharacterSet:2
    ].
    (ascii == 16rdc) ifTrue:[   "Udiaeresis"
	^ self character:$] fromCharacterSet:2
    ].
    (ascii == 16rc4) ifTrue:[   "Adiaeresis"
	^ self character:$[ fromCharacterSet:2
    ].
    (ascii == 16rd6) ifTrue:[   "Odiaeresis"
	^ self character:$\ fromCharacterSet:2
    ].
    (ascii == 16rdf) ifTrue:[   "ssharp"
	^ self character:$~ fromCharacterSet:2
    ]
!

normal
    "set font to normal"

    "send: <ESC>H<ESC>5"

    super escape:$H. super escape:$5
! !

!EpsonFX1PrinterStream methodsFor:'layout'!

linesPerPage:n
    "send: <ESC>C<n>"

    super escape:$C.
    super nextPut:(Character value:n)
!

newPage
    super nextPut:(Character value:12)
!

noTopMargin
    "send <ESC>0"

    super escape:$O
!

pageHeight:inch
    "send <ESC>G<0><inch>"

    super escape:$C.
    super nextPut:(Character value:0).
    super nextPut:(Character value:inch)
!

proportional:aBoolean
    "send <ESC>p0 or <ESC>p1"

    super escape:$p.
    aBoolean ifTrue:[
	super nextPut:$1
    ] ifFalse:[
	super nextPut:$0
    ]
!

reset
    "send <ESC>@"

    super escape:$@
!

topMargin:n
    "send <ESC>N<n>"

    super escape:$N.
    super nextPut:(Character value:n)
! !
